#!python
# Copyright 2023, Arcangelo Massari <arcangelo.massari@unibo.it>, Arianna Moretti <arianna.moretti4@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


import json
import os
from argparse import ArgumentParser
from functools import partial

import yaml
from pebble import ProcessFuture, ProcessPool
from rdflib import URIRef
from tqdm import tqdm

from oc_meta.plugins.editor import MetaEditor
from oc_meta.plugins.fixer.merge_duplicated_ids import (extract_identifiers,
                                                        process_archive)


def task_done(merge_index: dict, task_output:ProcessFuture) -> None:
    filepath, to_be_merged = task_output.result()
    merge_index[filepath.replace('\\', '/')] = to_be_merged
    if len(merge_index) % 1000 == 0:
        with open(CACHE_PATH, 'w', encoding='utf-8') as f:
            json.dump(merge_index, f)
    pbar.update()

if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('duplicated_ids.py', description='Merge duplicated ids for the entity type specified')
    arg_parser.add_argument('-e', '--entity_type', dest='entity_type', required=True, choices=['ra', 'br'], help='An entity type abbreviation')
    arg_parser.add_argument('-c', '--meta_config', dest='meta_config', required=True, help='OpenCitations Meta configuration file location')
    arg_parser.add_argument('-r', '--resp_agent', dest='resp_agent', required=True, help='Your ORCID URL')
    arg_parser.add_argument('-ca', '--cache', dest='cache', required=True, help='Cache filepath')
    args = arg_parser.parse_args()
    with open(args.meta_config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    rdf_dir = os.path.join(settings['output_rdf_dir'], 'rdf') + os.sep
    rdf_entity_dir = os.path.join(rdf_dir, args.entity_type) + os.sep
    zip_output_rdf = settings['zip_output_rdf']
    file_extension = '.zip' if zip_output_rdf else '.json'
    merge_index = {'already_merged': list()}
    CACHE_PATH = args.cache
    if os.path.exists(CACHE_PATH):
        with open(CACHE_PATH, 'r', encoding='utf-8') as f:
            merge_index: dict = json.load(f)
    filepaths = [os.path.join(fold, file) for fold, _, files in os.walk(rdf_entity_dir) for file in files 
        if file.endswith(file_extension) 
        and os.path.basename(fold) != 'prov'
        and os.path.join(fold, file).replace('\\', '/') not in merge_index]
    pbar = tqdm(total=len(filepaths))
    rdf_dir = os.path.join(settings['output_rdf_dir'], 'rdf') + os.sep
    dir_split_number = settings['dir_split_number']
    items_per_file = settings['items_per_file']
    meta_config = args.meta_config
    resp_agent = settings['resp_agent']
    memory = dict()
    for dirpath, _, filenames in os.walk(rdf_dir):
        for filename in filenames:
            if filename.endswith('.lock'):
                os.remove(os.path.join(dirpath, filename))
    with ProcessPool(max_tasks=1) as executor:
        for filepath in filepaths:
            future:ProcessFuture = executor.schedule(
                function=process_archive, 
                args=(filepath, extract_identifiers, memory, filepath, rdf_dir, dir_split_number, items_per_file, zip_output_rdf, memory, meta_config, resp_agent)) 
            future.add_done_callback(partial(task_done, merge_index))
    pbar.close()
    meta_editor = MetaEditor(meta_config, resp_agent)
    pbar = tqdm(len(merge_index))
    for filepath, couples in merge_index.items():
        if filepath != 'already_merged':
            for couple in couples:
                if couple[0] not in merge_index['already_merged'] and couple[1] not in merge_index['already_merged']:
                    meta_editor.merge(URIRef(couple[0]), URIRef(couple[1]))
                    merge_index['already_merged'].append(couple[1])
                    with open(CACHE_PATH, 'w', encoding='utf-8') as f:
                        json.dump(merge_index, f)
        pbar.update()
    pbar.close()


import os
import zipfile

def zip_individual_files_in_folder(folder_path):
    for foldername, subfolders, filenames in os.walk(folder_path):
        for filename in filenames:
            file_path = os.path.join(foldername, filename)
            zip_filename = os.path.join(foldername, f"{filename}.zip")
            
            with zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED) as zipf:
                zipf.write(file_path, os.path.basename(file_path))

if __name__ == "__main__":
    folder_to_zip  = '/srv/data/arcangelo/export_test/rdf/'  # Inserisci il percorso della cartella che vuoi zippare
    zip_individual_files_in_folder(folder_to_zip)
    print(f"File zippati individualmente nella cartella di origine.")


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2021-2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from argparse import ArgumentParser

from rdflib import URIRef

from oc_meta.plugins.editor import MetaEditor

if __name__ == '__main__':
    arg_parser = ArgumentParser('meta_editor.py', description='This script edits OpenCitations Meta triplestore, RDF and provenance')
    arg_parser.add_argument('-c', '--config', dest='config', required=True, help='Configuration file directory')
    arg_parser.add_argument('-op', '--operation', dest='operation', required=True, choices=['update', 'delete', 'sync', 'merge'], help='The CRUD operation to perform')
    arg_parser.add_argument('-s', '--subject', dest='res', required=True, type=URIRef, help='The subject entity')
    arg_parser.add_argument('-p', '--property', dest='property', required=False, help='The property')
    arg_parser.add_argument('-o', '--object', dest='value', required=False, help='The value')
    arg_parser.add_argument('-ot', '--other', dest='other', required=False, help='Other res to be merged with res')
    arg_parser.add_argument('-r', '--resp', dest='resp_agent', required=True, help='Your ORCID')
    args = arg_parser.parse_args()
    meta_editor = MetaEditor(args.config, args.resp_agent)
    if args.operation == 'update':
        meta_editor.update_property(args.res, args.property, args.value)
    elif args.operation == 'delete':
        meta_editor.delete(args.res, args.property, args.value)
    elif args.operation == 'sync':
        meta_editor.sync_rdf_with_triplestore(args.res)
    elif args.operation == 'merge':
        meta_editor.merge(URIRef(args.res), URIRef(args.other))


import argparse
import csv
from SPARQLWrapper import SPARQLWrapper, JSON

class UnionFind:
    def __init__(self):
        self.parent = {}
        self.rank = {}

    def make_set(self, x):
        if x not in self.parent:
            self.parent[x] = x
            self.rank[x] = 0

    def find(self, x):
        if x not in self.parent:
            self.make_set(x)
        if self.parent[x] != x:
            self.parent[x] = self.find(self.parent[x])  # Path compression
        return self.parent[x]

    def union(self, x, y):
        root_x = self.find(x)
        root_y = self.find(y)
        if root_x != root_y:
            if self.rank[root_x] < self.rank[root_y]:
                self.parent[root_x] = root_y
            elif self.rank[root_x] > self.rank[root_y]:
                self.parent[root_y] = root_x
            else:
                self.parent[root_y] = root_x
                self.rank[root_x] += 1

def find_surviving_entities(results):
    uf = UnionFind()
    for result in results["results"]["bindings"]:
        entity1 = result["entity1"]["value"]
        entity2 = result["entity2"]["value"]
        uf.union(entity1, entity2)

    final_entities = {}
    for entity in uf.parent:
        root = uf.find(entity)
        if root not in final_entities:
            final_entities[root] = set()
        if root != entity:
            final_entities[root].add(entity)

    return final_entities

def execute_sparql_query(endpoint_url, query):
    sparql = SPARQLWrapper(endpoint_url)
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    return sparql.query().convert()

def write_results_to_csv(final_entities, csv_file_path):
    with open(csv_file_path, mode='w', newline='', encoding='utf-8') as csv_file:
        writer = csv.writer(csv_file)
        writer.writerow(['surviving_entity', 'merged_entities'])

        for surviving_entity, merged_entities in final_entities.items():
            if merged_entities:  # Only write rows with actual merges
                writer.writerow([surviving_entity, '; '.join(merged_entities)])

    print(f"CSV file written to {csv_file_path}")

def get_sparql_query(entity_type):
    if entity_type == 'id':
        return """
            PREFIX datacite: <http://purl.org/spar/datacite/>
            PREFIX literal: <http://www.essepuntato.it/2010/06/literalreification/>
            SELECT ?entity1 ?entity2 {
                ?entity1 datacite:usesIdentifierScheme ?scheme;
                    literal:hasLiteralValue ?literal.
                ?entity2 datacite:usesIdentifierScheme ?scheme;
                    literal:hasLiteralValue ?literal.
                FILTER(?entity1 != ?entity2 )
            }    
        """
    elif entity_type == 'ra':
        return """
            PREFIX datacite: <http://purl.org/spar/datacite/>
            PREFIX literal: <http://www.essepuntato.it/2010/06/literalreification/>
            PREFIX foaf: <http://xmlns.com/foaf/0.1/>
            SELECT ?entity1 ?entity2 {
                ?entity1 datacite:hasIdentifier ?id;
                        a foaf:Agent.
                ?entity2 datacite:hasIdentifier ?id;
                        a foaf:Agent.
                FILTER(?entity1 != ?entity2 )
            }    
        """
    else:
        raise ValueError(f"Unsupported entity type: {entity_type}")

def debug_specific_entities(results):
    print("Debug: Results for specific entities")
    if not results["results"]["bindings"]:
        print("No results found for the specified entities.")
    for result in results["results"]["bindings"]:
        entity = result.get("entity", {}).get("value", "N/A")
        scheme = result.get("scheme", {}).get("value", "N/A")
        literal = result.get("literal", {}).get("value", "N/A")
        print(f"Entity: {entity}")
        print(f"  Scheme: {scheme}")
        print(f"  Literal: {literal}")
        print("---")

def get_debug_sparql_query():
    return """
    PREFIX datacite: <http://purl.org/spar/datacite/>
    PREFIX literal: <http://www.essepuntato.it/2010/06/literalreification/>

    SELECT ?entity ?scheme ?literal
    WHERE {
      VALUES ?entity {
        <https://w3id.org/oc/meta/id/061705640086>
        <https://w3id.org/oc/meta/id/061705582835>
      }
      OPTIONAL {
        ?entity datacite:usesIdentifierScheme ?scheme .
      }
      OPTIONAL {
        ?entity literal:hasLiteralValue ?literal .
      }
    }
    """

def main():
    parser = argparse.ArgumentParser(description='Run a SPARQL query and save results to CSV.')
    parser.add_argument('endpoint', type=str, help='SPARQL endpoint URL')
    parser.add_argument('csv_path', type=str, help='Path to the CSV file to save results')
    parser.add_argument('entity_type', type=str, choices=['id', 'br', 'ra'], help='Type of entity: "id", "br", or "ra"')
    args = parser.parse_args()

    sparql_query = get_sparql_query(args.entity_type)
    results = execute_sparql_query(args.endpoint, sparql_query)
    
    # sparql_query = get_debug_sparql_query()
    # results = execute_sparql_query(args.endpoint, sparql_query)

    # debug_specific_entities(results)
    
    final_entities = find_surviving_entities(results)
    write_results_to_csv(final_entities, args.csv_path)

if __name__ == '__main__':
    main()


import argparse
import logging
import logging.handlers
import os
import re
import zipfile
from datetime import datetime, timezone
from multiprocessing import cpu_count, current_process
from typing import List, Optional, Tuple
from zoneinfo import ZoneInfo

from pebble import ProcessPool
from rdflib import ConjunctiveGraph, Literal, Namespace, URIRef
from rdflib.namespace import RDF, XSD
from tqdm import tqdm

PROV = Namespace("http://www.w3.org/ns/prov#")

class TqdmToLogger:
    """
    Output stream for tqdm which will output to logger.
    """
    def __init__(self, logger, level=logging.INFO):
        self.logger = logger
        self.level = level
        self.last_msg = ''

    def write(self, buf):
        # Don't print empty lines
        msg = buf.strip()
        if msg and msg != self.last_msg:
            self.logger.log(self.level, msg)
            self.last_msg = msg
    
    def flush(self):
        pass

class LazyLogger:
    """Logger that only creates file when first log message is written."""
    def __init__(self, log_dir: str, logger_name: str):
        self.log_dir = log_dir
        self.logger_name = logger_name
        self.logger = None
        self.process_name = current_process().name
        self.timestamp = None
        self._has_logged = False

    def _initialize_logger(self):
        if not self.logger:
            os.makedirs(self.log_dir, exist_ok=True)
            self.timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            log_filename = f"{self.logger_name}_{self.process_name}_{self.timestamp}.log"
            
            self.logger = logging.getLogger(f"{self.logger_name}_{self.process_name}_{self.timestamp}")
            self.logger.setLevel(logging.INFO)
            
            handler = logging.FileHandler(
                filename=os.path.join(self.log_dir, log_filename),
                encoding='utf-8',
                mode='a'
            )
            
            formatter = logging.Formatter('%(asctime)s - %(message)s')
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.propagate = False

    def log(self, level: int, message: str):
        if not self._has_logged:
            self._initialize_logger()
            self._has_logged = True
        self.logger.log(level, message)

    def info(self, message: str):
        self.log(logging.INFO, message)

    def error(self, message: str):
        self.log(logging.ERROR, message)

    def warning(self, message: str):
        self.log(logging.WARNING, message)

    def has_logged(self) -> bool:
        return self._has_logged

class ProvenanceProcessor:
    def __init__(self, log_dir: str):
        self._snapshot_number_pattern = re.compile(r'/prov/se/(\d+)$')
        self._default_time = Literal(
            datetime(2022, 12, 20, tzinfo=timezone.utc).isoformat(),
            datatype=XSD.dateTime
        )
        self.log_dir = log_dir
        self.logger = LazyLogger(log_dir, 'modifications')

    def _log_modification(self, entity_uri: str, mod_type: str, message: str) -> None:
        self.logger.info(f"Entity: {entity_uri} - {mod_type} - {message}")

    def _extract_snapshot_number(self, snapshot_uri: str) -> int:
        """Extract the snapshot number from its URI using pre-compiled regex."""
        match = self._snapshot_number_pattern.search(str(snapshot_uri))
        return int(match.group(1)) if match else 0

    def _get_entity_from_prov_graph(self, graph_uri: str) -> str:
        """Extract entity URI from its provenance graph URI."""
        return str(graph_uri).replace('/prov/', '')

    def _convert_to_utc(self, timestamp_str: str) -> datetime:
        """Convert a timestamp string to UTC datetime."""
        timestamp_str = str(timestamp_str).replace('Z', '+00:00')
        dt = datetime.fromisoformat(timestamp_str)
        
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=ZoneInfo("Europe/Rome"))
        
        return dt.astimezone(timezone.utc)

    def _normalize_timestamp(self, literal: Literal) -> Tuple[Literal, bool]:
        """Normalize a timestamp literal to UTC timezone."""
        try:
            dt = self._convert_to_utc(literal)
            new_literal = Literal(dt.isoformat(), datatype=XSD.dateTime)
            return new_literal, str(new_literal) != str(literal)
        except Exception as e:
            self.logger.error(f"Error normalizing timestamp {literal}: {e}")
            return literal, False

    def _collect_snapshot_info(self, context: ConjunctiveGraph) -> List[dict]:
        """Collect all snapshot information in a single pass."""
        snapshots = []
        seen_uris = set()
        base_uri = None
        
        for s, p, o in context.triples((None, None, None)):
            if '/prov/se/' in str(s) and str(s) not in seen_uris:
                if base_uri is None:
                    base_uri = str(s).rsplit('/se/', 1)[0]
                
                snapshot = {
                    'uri': s,
                    'number': self._extract_snapshot_number(s),
                    'generation_times': list(context.objects(s, PROV.generatedAtTime, unique=True)),
                    'invalidation_times': list(context.objects(s, PROV.invalidatedAtTime, unique=True))
                }
                snapshots.append(snapshot)
                seen_uris.add(str(s))
        
        sorted_snapshots = sorted(snapshots, key=lambda x: x['number'])
        
        if len(sorted_snapshots) >= 2:
            # Identify and fill missing snapshots
            filled_snapshots = self._fill_missing_snapshots(context, sorted_snapshots, base_uri)
            return filled_snapshots
            
        return sorted_snapshots

    def _fill_missing_snapshots(self, context: ConjunctiveGraph, 
                                snapshots: List[dict], base_uri: str) -> List[dict]:
        """Fill in missing snapshots in the sequence."""
        if not snapshots:
            return snapshots
            
        filled_snapshots = []
        max_num = max(s['number'] for s in snapshots)
        min_num = min(s['number'] for s in snapshots)
        existing_numbers = {s['number'] for s in snapshots}
        existing_snapshots = {s['number']: s for s in snapshots}
        
        start_num = 1
        
        # Itera da 1 fino al massimo numero di snapshot
        for i in range(start_num, max_num + 1):
            if i in existing_numbers:
                filled_snapshots.append(existing_snapshots[i])
            else:
                # Create missing snapshot
                missing_uri = URIRef(f"{base_uri}/se/{i}")
                
                # Find the first previous available snapshot
                prev_num = i - 1
                while prev_num >= start_num and prev_num not in existing_numbers:
                    prev_num -= 1
                prev_snapshot = existing_snapshots.get(prev_num)
                
                # Find the first next available snapshot
                next_num = i + 1
                while next_num <= max_num and next_num not in existing_numbers:
                    next_num += 1
                next_snapshot = existing_snapshots.get(next_num)
                
                missing_snapshot = self._create_missing_snapshot(
                    context, missing_uri, i, 
                    prev_snapshot, next_snapshot
                )
                
                # Add the missing snapshot to existing_snapshots and existing_numbers
                existing_snapshots[i] = missing_snapshot
                existing_numbers.add(i)
                filled_snapshots.append(missing_snapshot)
                    
        return sorted(filled_snapshots, key=lambda x: x['number'])

    def _create_missing_snapshot(self, context: ConjunctiveGraph, 
                            missing_uri: URIRef, number: int,
                            prev_snapshot: Optional[dict], 
                            next_snapshot: Optional[dict]) -> dict:
        """Create a missing snapshot with basic information."""
        entity_uri = URIRef(self._get_entity_from_prov_graph(str(missing_uri).split('se')[0]))
        
        # Add basic triples for the missing snapshot
        context.add((missing_uri, RDF.type, PROV.Entity))
        context.add((missing_uri, PROV.specializationOf, entity_uri))
        
        # Add wasDerivedFrom if we have a previous snapshot
        if prev_snapshot:
            context.add((missing_uri, PROV.wasDerivedFrom, prev_snapshot['uri']))
        
        # Se è lo snapshot 1, aggiungi la descrizione di creazione
        if number == 1:
            description = Literal(f"The entity '{str(entity_uri)}' has been created.")
            context.add((
                missing_uri, 
                URIRef("http://purl.org/dc/terms/description"), 
                description
            ))
            self._log_modification(
                str(missing_uri),
                "Added creation description for first snapshot",
                str(description)
            )
        
        generation_time = None
        invalidation_time = None
        
        # First check if it is the first snapshot and lacks generation time
        if not prev_snapshot and number == 1:
            generation_time = self._default_time
        else:
            # If we have previous snapshot, generation time is its invalidation time
            if prev_snapshot and prev_snapshot['invalidation_times']:
                if number == prev_snapshot['number'] + 1:
                    generation_time = prev_snapshot['invalidation_times'][0]
            
            # If we have next snapshot, invalidation time is its generation time
            if next_snapshot and next_snapshot['generation_times']:
                if number == next_snapshot['number'] - 1:
                    invalidation_time = next_snapshot['generation_times'][0]
                    
            # If both times are missing, or one of them, we need to compute intermediate times
            if not generation_time or not invalidation_time:
                start_time = None
                end_time = None
                
                # Find temporal reference points
                if prev_snapshot:
                    if prev_snapshot['invalidation_times']:
                        start_time = self._convert_to_utc(prev_snapshot['invalidation_times'][0])
                    elif prev_snapshot['generation_times']:
                        start_time = self._convert_to_utc(prev_snapshot['generation_times'][0])
                        
                if next_snapshot:
                    if next_snapshot['generation_times']:
                        end_time = self._convert_to_utc(next_snapshot['generation_times'][0])
                    elif next_snapshot['invalidation_times']:
                        end_time = self._convert_to_utc(next_snapshot['invalidation_times'][0])
                
                if start_time and end_time:
                    # Calculate how many intervals are between existing snapshots
                    total_missing = next_snapshot['number'] - prev_snapshot['number'] - 1
                    position = number - prev_snapshot['number']
                    
                    # Uniformly distribute missing times
                    interval = (end_time - start_time) / (total_missing + 1)
                    intermediate_time = start_time + (interval * position)
                    
                    # Use computed times only if we lack the real ones
                    if not generation_time:
                        generation_time = Literal(intermediate_time.isoformat(), datatype=XSD.dateTime)
                    if not invalidation_time:
                        next_time = intermediate_time + interval
                        invalidation_time = Literal(next_time.isoformat(), datatype=XSD.dateTime)
                        
        # Add timestamps
        if generation_time:
            context.add((missing_uri, PROV.generatedAtTime, generation_time))
            self._log_modification(
                str(missing_uri),
                "Added generatedAtTime for missing snapshot",
                str(generation_time)
            )
            
        if invalidation_time:
            context.add((missing_uri, PROV.invalidatedAtTime, invalidation_time))
            self._log_modification(
                str(missing_uri),
                "Added invalidatedAtTime for missing snapshot",
                str(invalidation_time)
            )
            
        self._log_modification(
            str(missing_uri),
            "Created missing snapshot",
            f"number {number}"
        )
        
        return {
            'uri': missing_uri,
            'number': number,
            'generation_times': [generation_time] if generation_time else [],
            'invalidation_times': [invalidation_time] if invalidation_time else []
        }

    def _remove_multiple_timestamps(self, context: ConjunctiveGraph, 
                                  snapshot_uri: URIRef, 
                                  predicate: URIRef, 
                                  timestamps: List[Literal]) -> None:
        """Remove all timestamps for a given predicate."""
        for ts in timestamps:
            context.remove((snapshot_uri, predicate, ts))
            self._log_modification(
                str(snapshot_uri),
                f"Removed {predicate.split('#')[-1]}",
                f"{str(ts)}"
            )

    @staticmethod
    def process_file(prov_file_path: str, log_dir: str) -> Optional[bool]:
        process_logger = LazyLogger(log_dir, 'processing')
        processor = ProvenanceProcessor(log_dir)
        
        try:
            with zipfile.ZipFile(prov_file_path, 'r') as zip_ref:
                g = ConjunctiveGraph()
                for filename in zip_ref.namelist():
                    with zip_ref.open(filename) as file:
                        g.parse(file, format='json-ld')
                
                modified = False
                
                for context in g.contexts():
                    context_uri = str(context.identifier)
                    if not context_uri.endswith('/prov/'):
                        continue
                    
                    entity_uri = URIRef(processor._get_entity_from_prov_graph(context_uri))
                    snapshots = processor._collect_snapshot_info(context)
                    
                    if snapshots:
                        modified |= processor._process_snapshots(context, entity_uri, snapshots)

                if modified:
                    with zipfile.ZipFile(prov_file_path, 'w', zipfile.ZIP_DEFLATED, allowZip64=True) as zip_out:
                        jsonld_data = g.serialize(format='json-ld', encoding='utf-8', 
                                               ensure_ascii=False, indent=None)
                        zip_out.writestr('se.json', jsonld_data)
                    
                    return True
                
                return False
                
        except Exception as e:
            process_logger.error(f"Error processing {prov_file_path}: {e}")
            return None

    def _process_snapshots(self, context: ConjunctiveGraph, entity_uri: URIRef, 
                        snapshots: List[dict]) -> bool:
        """Process all snapshots in batch operations."""
        modified = False

        # Handle multiple descriptions for each snapshot
        for i, snapshot in enumerate(snapshots):
            is_first = (i == 0)
            is_last = (i == len(snapshots) - 1)
            modified |= self._handle_multiple_descriptions(context, snapshot['uri'], is_first, is_last)

        # Process specializationOf relationships
        for snapshot in snapshots:
            if not any(context.objects(snapshot['uri'], PROV.specializationOf)):
                context.add((snapshot['uri'], PROV.specializationOf, entity_uri))
                self._log_modification(
                    str(entity_uri), 
                    "Added specializationOf",
                    str(snapshot['uri'])
                )
                modified = True

        # Process wasDerivedFrom relationships
        for i in range(1, len(snapshots)):
            curr_snapshot = snapshots[i]
            prev_snapshot = snapshots[i-1]
            
            # Check if current snapshot is a merge snapshot
            curr_desc = list(context.objects(curr_snapshot['uri'], 
                                        URIRef("http://purl.org/dc/terms/description")))
            is_merge = any("has been merged with" in str(desc) for desc in curr_desc)
            
            derived_from = list(context.objects(curr_snapshot['uri'], PROV.wasDerivedFrom))

            if is_merge:
                # For merge snapshots, ensure there's at least the wasDerivedFrom link to the previous snapshot
                if not derived_from or prev_snapshot['uri'] not in derived_from:
                    context.add((curr_snapshot['uri'], PROV.wasDerivedFrom, prev_snapshot['uri']))
                    self._log_modification(
                        str(entity_uri),
                        "Added merge wasDerivedFrom",
                        f"{str(curr_snapshot['uri'])} → {str(prev_snapshot['uri'])}"
                    )
                    modified = True
            else:
                # For non-merge snapshots, ensure exactly one wasDerivedFrom link to previous snapshot
                if len(derived_from) != 1 or derived_from[0] != prev_snapshot['uri']:
                    # Remove any existing wasDerivedFrom relations
                    for df in derived_from:
                        context.remove((curr_snapshot['uri'], PROV.wasDerivedFrom, df))
                    
                    # Add the correct wasDerivedFrom relation
                    context.add((curr_snapshot['uri'], PROV.wasDerivedFrom, prev_snapshot['uri']))
                    self._log_modification(
                        str(entity_uri),
                        "Fixed wasDerivedFrom",
                        f"{str(curr_snapshot['uri'])} → {str(prev_snapshot['uri'])}"
                    )
                    modified = True

        # Process timestamps
        modified |= self._process_timestamps(context, snapshots)
        
        return modified

    def _process_timestamps(self, context: ConjunctiveGraph, snapshots: List[dict]) -> bool:
        """Process all timestamps in batch."""
        modified = False
        
        for i, snapshot in enumerate(snapshots):
            # Handle generation time
            modified |= self._handle_generation_time(context, snapshots, i)
            
            # Handle invalidation time
            if i < len(snapshots) - 1:
                modified |= self._handle_invalidation_time(context, snapshots, i)
        
        return modified

    def _get_earliest_timestamp(self, timestamps: List[Literal]) -> Literal:
        """Return the earliest timestamp from a list of timestamps."""
        earliest = min(
            self._convert_to_utc(ts) 
            for ts in timestamps
        )
        return Literal(earliest.isoformat(), datatype=XSD.dateTime)

    def _handle_generation_time(self, context: ConjunctiveGraph, 
                            snapshots: List[dict], index: int) -> bool:
        """Handle generation time for a snapshot."""
        modified = False
        snapshot = snapshots[index]

        if len(snapshot['generation_times']) > 1:
            # Se ci sono multipli timestamp, mantieni solo il più vecchio
            earliest_time = self._get_earliest_timestamp(snapshot['generation_times'])
            
            # Rimuovi tutti i timestamp esistenti
            self._remove_multiple_timestamps(
                context, snapshot['uri'], PROV.generatedAtTime, snapshot['generation_times']
            )
            
            # Aggiungi il timestamp più vecchio
            context.add((snapshot['uri'], PROV.generatedAtTime, earliest_time))
            self._log_modification(
                str(snapshot['uri']),
                "Kept earliest generatedAtTime",
                f"{str(earliest_time)}"
            )
            modified = True
        elif len(snapshot['generation_times']) == 0:
            new_time = None
            
            if index > 0:
                prev_snapshot = snapshots[index-1]
                if prev_snapshot['invalidation_times']:
                    new_time = prev_snapshot['invalidation_times'][0]
                elif (prev_snapshot['generation_times'] and
                    snapshot['invalidation_times'] and 
                    len(snapshot['invalidation_times']) == 1):
                    prev_time = self._convert_to_utc(prev_snapshot['generation_times'][0])
                    curr_time = self._convert_to_utc(snapshot['invalidation_times'][0])
                    middle_time = prev_time + (curr_time - prev_time) / 2
                    new_time = Literal(middle_time.isoformat(), datatype=XSD.dateTime)
            else:
                new_time = self._default_time
            
            if new_time:
                context.add((snapshot['uri'], PROV.generatedAtTime, new_time))
                self._log_modification(
                    str(snapshot['uri']),
                    "Added generatedAtTime",
                    f"{str(new_time)}"
                )
                modified = True
        
        return modified

    def _handle_invalidation_time(self, context: ConjunctiveGraph, 
                                snapshots: List[dict], index: int) -> bool:
        """Handle invalidation time for a snapshot.
        
        When both current snapshot's invalidation time and next snapshot's generation time
        are missing, calculates an intermediate time based on available timestamps and 
        ensures synchronization between snapshots.
        """
        modified = False
        snapshot = snapshots[index]
        next_snapshot = snapshots[index + 1]

        if len(snapshot['invalidation_times']) > 1:
            # Se ci sono multipli timestamp, mantieni solo il più vecchio
            earliest_time = self._get_earliest_timestamp(snapshot['invalidation_times'])
            
            # Rimuovi tutti i timestamp esistenti
            self._remove_multiple_timestamps(
                context, snapshot['uri'], PROV.invalidatedAtTime, 
                snapshot['invalidation_times']
            )
            
            # Aggiungi il timestamp più vecchio
            context.add((snapshot['uri'], PROV.invalidatedAtTime, earliest_time))
            self._log_modification(
                str(snapshot['uri']),
                "Kept earliest invalidatedAtTime",
                f"{str(earliest_time)}"
            )
            modified = True
        elif len(snapshot['invalidation_times']) == 0:
            new_time = None
            
            if next_snapshot['generation_times']:
                # Se lo snapshot successivo ha un tempo di generazione, usalo
                if len(next_snapshot['generation_times']) == 1:
                    new_time = next_snapshot['generation_times'][0]
                else:
                    earliest_time = self._get_earliest_timestamp(next_snapshot['generation_times'])
                    new_time = earliest_time
            else:
                # Caso in cui mancano entrambi i tempi
                # Cerchiamo di calcolare un tempo intermedio basandoci sui tempi disponibili
                if snapshot['generation_times'] and next_snapshot['invalidation_times']:
                    curr_gen_time = self._convert_to_utc(snapshot['generation_times'][0])
                    next_inv_time = self._convert_to_utc(next_snapshot['invalidation_times'][0])
                    
                    # Calcola un tempo intermedio
                    intermediate_time = curr_gen_time + (next_inv_time - curr_gen_time) / 2
                    new_time = Literal(intermediate_time.isoformat(), datatype=XSD.dateTime)
                    
                    # Importante: aggiungi lo stesso tempo come generatedAtTime allo snapshot successivo
                    context.add((next_snapshot['uri'], PROV.generatedAtTime, new_time))
                    self._log_modification(
                        str(next_snapshot['uri']),
                        "Added synchronized generatedAtTime",
                        f"{str(new_time)}"
                    )
                    modified = True

            if new_time:
                context.add((snapshot['uri'], PROV.invalidatedAtTime, new_time))
                self._log_modification(
                    str(snapshot['uri']),
                    "Added invalidatedAtTime",
                    f"{str(new_time)}"
                )
                modified = True
        
        return modified

    def _handle_multiple_descriptions(self, context: ConjunctiveGraph, snapshot_uri: URIRef, 
                                is_first_snapshot: bool, is_last_snapshot: bool) -> bool:
        """Handle cases where a snapshot has multiple descriptions.
        
        Args:
            context (ConjunctiveGraph): The RDF graph context
            snapshot_uri (URIRef): The URI of the snapshot to process
            is_first_snapshot (bool): Whether this is the first snapshot in the sequence
            is_last_snapshot (bool): Whether this is the last snapshot in the sequence
            
        Returns:
            bool: True if modifications were made, False otherwise
        """
        modified = False
        descriptions = list(context.objects(snapshot_uri, URIRef("http://purl.org/dc/terms/description")))
        
        if len(descriptions) <= 1:
            return modified
            
        # Check for merge descriptions
        merge_descriptions = [desc for desc in descriptions if "has been merged with" in str(desc)]

        if merge_descriptions and not is_first_snapshot:
            # For merge snapshots (not being the first one), keep all merge descriptions
            # and remove any non-merge descriptions
            for desc in descriptions:
                if "has been merged with" not in str(desc):
                    context.remove((snapshot_uri, URIRef("http://purl.org/dc/terms/description"), desc))
                    self._log_modification(
                        str(snapshot_uri),
                        "Removed non-merge description from merge snapshot",
                        str(desc)
                    )
                    modified = True
        else:
            # For non-merge cases, apply lifecycle-based priority
            selected_desc = None
            
            if is_first_snapshot:
                # First snapshot must be creation
                creation_descs = [desc for desc in descriptions if "has been created" in str(desc)]
                if creation_descs:
                    selected_desc = creation_descs[0]
            elif is_last_snapshot:
                # Last snapshot could be deletion
                deletion_descs = [desc for desc in descriptions if "has been deleted" in str(desc)]
                if deletion_descs:
                    selected_desc = deletion_descs[0]
            
            # If no specific lifecycle description was selected, default to modification
            if not selected_desc:
                modification_descs = [desc for desc in descriptions if "has been modified" in str(desc)]
                if modification_descs:
                    selected_desc = modification_descs[0]
            
            if selected_desc:
                # Remove all descriptions except the selected one
                for desc in descriptions:
                    if desc != selected_desc:
                        context.remove((snapshot_uri, URIRef("http://purl.org/dc/terms/description"), desc))
                        self._log_modification(
                            str(snapshot_uri),
                            "Removed duplicate description based on lifecycle position",
                            str(desc)
                        )
                        modified = True
        
        return modified

def process_chunk(args):
    file_chunk, log_dir = args
    results = []
    chunk_logger = LazyLogger(log_dir, 'chunk')
    
    for file in file_chunk:
        result = ProvenanceProcessor.process_file(file, log_dir)
        if result:
            results.append(file)
            if result is True:  # Only log if modifications were made
                chunk_logger.info(f"Modified file: {file}")
    
    return results

def main():
    parser = argparse.ArgumentParser(description="Fix provenance files in parallel")
    parser.add_argument('input_dir', type=str, help="Directory containing provenance files")
    parser.add_argument('--processes', type=int, default=cpu_count(),
                       help="Number of parallel processes (default: number of CPU cores)")
    parser.add_argument('--log-dir', type=str, default='logs',
                       help="Directory for log files (default: logs)")
    args = parser.parse_args()
    
    os.makedirs(args.log_dir, exist_ok=True)
    main_logger = LazyLogger(args.log_dir, 'main')
    
    main_logger.info("Starting provenance fix process")
    
    prov_files = [
        os.path.join(root, file)
        for root, _, files in os.walk(args.input_dir)
        for file in files
        if file.endswith('se.zip')
    ]

    chunk_size = 100
    file_chunks = [prov_files[i:i + chunk_size] 
                  for i in range(0, len(prov_files), chunk_size)]
    
    chunk_args = [(chunk, args.log_dir) for chunk in file_chunks]

    modified_files = []
    process_logger = LazyLogger(args.log_dir, 'processing')
    tqdm_output = TqdmToLogger(process_logger)

    with ProcessPool(max_workers=args.processes, max_tasks=1) as pool:
        future = pool.map(process_chunk, chunk_args)
        iterator = future.result()

        with tqdm(
            total=len(file_chunks), 
            desc="Fixing provenance files", 
            file=tqdm_output,
            mininterval=5.0,
            maxinterval=10.0) as pbar:
            try:
                while True:
                    try:
                        result = next(iterator)
                        modified_files.extend(result)
                        pbar.update(1)
                    except StopIteration:
                        break
                    except TimeoutError as error:
                        main_logger.error(f"Chunk processing timed out: {error}")
                    except Exception as error:
                        main_logger.error(f"Error processing chunk: {error}")
            except KeyboardInterrupt:
                main_logger.warning("Processing interrupted by user")
                future.cancel()
                raise

    if modified_files:
        main_logger.info(f"Process completed. Total files modified: {len(modified_files)}")
        print(f"\nProcessing complete. Check logs in {args.log_dir} directory.")
        print(f"Total files modified: {len(modified_files)}")
    else:
        print("\nProcessing complete. No files were modified.")

if __name__ == "__main__":
    main()


import argparse
import json
from oc_meta.plugins.editor import MetaEditor
from tqdm import tqdm
import os
from concurrent.futures import ProcessPoolExecutor, as_completed


def delete_uri(uri, meta_config, resp_agent, stop_file):
    if os.path.exists(stop_file):
        return "stopped"
    
    meta_editor = MetaEditor(meta_config=meta_config, resp_agent=resp_agent)
    print(f"Deleting entity: {uri}")
    meta_editor.delete(res=uri, property=None, object=None)
    return "deleted"


def single_process_deletion(uris, meta_config, resp_agent, stop_file):
    meta_editor = MetaEditor(meta_config=meta_config, resp_agent=resp_agent)
    
    with tqdm(total=len(uris), desc="Deleting URIs") as pbar:
        for uri in uris:
            if os.path.exists(stop_file):
                print("Stop file detected. Halting further deletions.")
                break
            print(f"Deleting entity: {uri}")
            meta_editor.delete(res=uri, property=None, object=None)
            pbar.update(1)


def multi_process_deletion(uris, meta_config, resp_agent, stop_file):
    with ProcessPoolExecutor() as executor:
        futures = {executor.submit(delete_uri, uri, meta_config, resp_agent, stop_file): uri for uri in uris}
        
        with tqdm(total=len(uris), desc="Deleting URIs") as pbar:
            for future in as_completed(futures):
                result = future.result()
                if result == "stopped":
                    print("Stop file detected. Halting further deletions.")
                    break
                pbar.update(1)


def main():
    parser = argparse.ArgumentParser(description="Delete URIs using MetaEditor.")
    parser.add_argument("json_file", type=str, help="Path to the JSON file containing URIs")
    parser.add_argument("meta_config", type=str, help="Path to the MetaEditor configuration file")
    parser.add_argument("resp_agent", type=str, help="Responsible agent URI")
    parser.add_argument("--stop_file", type=str, default=".stop_deletions", help="Path to the stop file for graceful termination")
    parser.add_argument("--multiprocessing", action="store_true", help="Use multiprocessing for deletions")

    args = parser.parse_args()

    if os.path.exists(args.stop_file):
        os.remove(args.stop_file)

    with open(args.json_file, 'r') as file:
        uris = json.load(file)

    if args.multiprocessing:
        multi_process_deletion(uris, args.meta_config, args.resp_agent, args.stop_file)
    else:
        single_process_deletion(uris, args.meta_config, args.resp_agent, args.stop_file)


if __name__ == "__main__":
    main()


import argparse
import os
import gzip
import re
from typing import Match
from rdflib import ConjunctiveGraph, URIRef, Literal, XSD
from multiprocessing import Pool, Manager
from tqdm import tqdm
import json
import zipfile
import shutil
from filelock import FileLock
from dateutil import parser as dateparser

# Definitions of regex patterns for parsing IRIs
entity_regex: str = r"^(.+)/([a-z][a-z])/(0[1-9]+0)?((?:[1-9][0-9]*)|(?:\d+-\d+))$"
prov_regex: str = r"^(.+)/([a-z][a-z])/(0[1-9]+0)?((?:[1-9][0-9]*)|(?:\d+-\d+))/prov/([a-z][a-z])/([1-9][0-9]*)$"

BASE_IRI = "https://w3id.org/oc/meta/"
DIR_SPLIT = 10000
N_FILE_ITEM = 1000
IS_JSON = True
PUBLICATION_DATE_PREDICATE = URIRef("http://prismstandard.org/namespaces/basic/2.0/publicationDate")

def _get_match(regex: str, group: int, string: str) -> str:
    match: Match = re.match(regex, string)
    if match is not None:
        return match.group(group)
    else:
        return ""

def get_base_iri(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return _get_match(prov_regex, 1, string_iri)
    else:
        return _get_match(entity_regex, 1, string_iri)

def get_short_name(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return _get_match(prov_regex, 5, string_iri)
    else:
        return _get_match(entity_regex, 2, string_iri)

def get_prov_subject_short_name(prov_res: URIRef) -> str:
    string_iri: str = str(prov_res)
    if "/prov/" in string_iri:
        return _get_match(prov_regex, 2, string_iri)
    else:
        return ""  # non-provenance entities do not have a prov_subject!

def get_prefix(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return ""  # provenance entities cannot have a supplier prefix
    else:
        return _get_match(entity_regex, 3, string_iri)

def get_prov_subject_prefix(prov_res: URIRef) -> str:
    string_iri: str = str(prov_res)
    if "/prov/" in string_iri:
        return _get_match(prov_regex, 3, string_iri)
    else:
        return ""  # non-provenance entities do not have a prov_subject!

def get_count(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return _get_match(prov_regex, 6, string_iri)
    else:
        return _get_match(entity_regex, 4, string_iri)

def get_prov_subject_count(prov_res: URIRef) -> str:
    string_iri: str = str(prov_res)
    if "/prov/" in string_iri:
        return _get_match(prov_regex, 4, string_iri)
    else:
        return ""  # non-provenance entities do not have a prov_subject!

def get_resource_number(res: URIRef) -> int:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return int(_get_match(prov_regex, 4, string_iri))
    else:
        return int(_get_match(entity_regex, 4, string_iri))

def find_local_line_id(res: URIRef, n_file_item: int = 1) -> int:
    cur_number: int = get_resource_number(res)

    cur_file_split: int = 0
    while True:
        if cur_number > cur_file_split:
            cur_file_split += n_file_item
        else:
            cur_file_split -= n_file_item
            break

    return cur_number - cur_file_split

def find_paths(res: URIRef, base_dir: str, base_iri: str, default_dir: str, dir_split: int,
               n_file_item: int, is_json: bool = True):
    """
    This function is responsible for looking for the correct JSON file that contains the data related to the
    resource identified by the variable 'string_iri'. This search takes into account the organisation in
    directories and files, as well as the particular supplier prefix for bibliographic entities, if specified.
    In case no supplier prefix is specified, the 'default_dir' (usually set to "_") is used instead.
    """
    string_iri: str = str(res)

    cur_number: int = get_resource_number(res)

    # Find the correct file number where to save the resources
    cur_file_split: int = 0
    while True:
        if cur_number > cur_file_split:
            cur_file_split += n_file_item
        else:
            break

    # The data have been split in multiple directories and it is not something related
    # with the provenance data of the whole corpus (e.g. provenance agents)
    if dir_split and not string_iri.startswith(base_iri + "prov/"):
        # Find the correct directory number where to save the file
        cur_split: int = 0
        while True:
            if cur_number > cur_split:
                cur_split += dir_split
            else:
                break

        if "/prov/" in string_iri:  # provenance file of a bibliographic entity
            subj_short_name: str = get_prov_subject_short_name(res)
            short_name: str = get_short_name(res)
            sub_folder: str = get_prov_subject_prefix(res)
            file_extension: str = '.json' if is_json else '.nq'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + subj_short_name + os.sep + sub_folder + \
                os.sep + str(cur_split) + os.sep + str(cur_file_split) + os.sep + "prov"
            cur_file_path: str = cur_dir_path + os.sep + short_name + file_extension
        else:  # regular bibliographic entity
            short_name: str = get_short_name(res)
            sub_folder: str = get_prefix(res)
            file_extension: str = '.json' if is_json else '.nt'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + short_name + os.sep + sub_folder + os.sep + str(cur_split)
            cur_file_path: str = cur_dir_path + os.sep + str(cur_file_split) + file_extension
    # Enter here if no split is needed
    elif dir_split == 0:
        if "/prov/" in string_iri:
            subj_short_name: str = get_prov_subject_short_name(res)
            short_name: str = get_short_name(res)
            sub_folder: str = get_prov_subject_prefix(res)
            file_extension: str = '.json' if is_json else '.nq'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + subj_short_name + os.sep + sub_folder + \
                os.sep + str(cur_file_split) + os.sep + "prov"
            cur_file_path: str = cur_dir_path + os.sep + short_name + file_extension
        else:
            short_name: str = get_short_name(res)
            sub_folder: str = get_prefix(res)
            file_extension: str = '.json' if is_json else '.nt'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + short_name + os.sep + sub_folder
            cur_file_path: str = cur_dir_path + os.sep + str(cur_file_split) + file_extension
    # Enter here if the data is about a provenance agent, e.g. /corpus/prov/
    else:
        short_name: str = get_short_name(res)
        prefix: str = get_prefix(res)
        count: str = get_count(res)
        file_extension: str = '.json' if is_json else '.nq'

        cur_dir_path: str = base_dir + short_name
        cur_file_path: str = cur_dir_path + os.sep + prefix + count + file_extension

    return cur_dir_path, cur_file_path

def log_error(log_dir, error_type, message, process_id):
    log_file_path = os.path.join(log_dir, f"log_{process_id}.json")
    if os.path.exists(log_file_path):
        with open(log_file_path, 'r', encoding='utf-8') as log_file:
            existing_log = json.load(log_file)
    else:
        existing_log = {}

    if error_type not in existing_log:
        existing_log[error_type] = []
    existing_log[error_type].append(message)

    with open(log_file_path, 'w', encoding='utf-8') as log_file:
        json.dump(existing_log, log_file, indent=4)

def merge_logs(log_dir, final_log_file):
    error_counts = {}
    final_log = {}

    for log_file_name in os.listdir(log_dir):
        if log_file_name.startswith("log_") and log_file_name.endswith(".json"):
            with open(os.path.join(log_dir, log_file_name), 'r', encoding='utf-8') as log_file:
                log_contents = json.load(log_file)
                for error_type, messages in log_contents.items():
                    if error_type not in final_log:
                        final_log[error_type] = []
                    final_log[error_type].extend(messages)

    for error_type, messages in final_log.items():
        error_counts[error_type] = len(messages)

    with open(final_log_file, 'w', encoding='utf-8') as flf:
        json.dump(final_log, flf, indent=4)

    summary_file_path = final_log_file.rsplit('.', 1)[0] + '_summary.json'
    with open(summary_file_path, 'w', encoding='utf-8') as sf:
        json.dump(error_counts, sf, indent=4)

    for log_file_name in os.listdir(log_dir):
        os.remove(os.path.join(log_dir, log_file_name))

def process_json_content(file_content, output_folder, zipped_output):
    subject_to_path_map = {}
    graph = ConjunctiveGraph()
    graph.parse(data=file_content, format='json-ld')
    for subject in graph.subjects(unique=True):
        _, cur_file_path = find_paths(subject, output_folder, BASE_IRI, "_", DIR_SPLIT, N_FILE_ITEM, IS_JSON)
        if zipped_output:
            cur_file_path = cur_file_path.replace('.json', '.zip')
        subject_to_path_map.setdefault(cur_file_path, []).append(subject)
    return graph, subject_to_path_map

def process_file(file_path, output_folder, log_dir, zipped_input, zipped_output, process_id, use_multiprocessing):
    if zipped_input:
        with zipfile.ZipFile(file_path, 'r') as zfile:
            for json_file_name in zfile.namelist():
                with zfile.open(json_file_name) as file:
                    file_content = file.read().decode('utf-8')
                    graph, subject_to_path_map = process_json_content(file_content, output_folder, zipped_output)
    else:
        with open(file_path, 'r', encoding='utf-8') as file:
            file_content = file.read()
            graph, subject_to_path_map = process_json_content(file_content, output_folder, zipped_output)
    
    for cur_file_path, subjects in subject_to_path_map.items():
        if not os.path.exists(cur_file_path):
            error_message = f"File {cur_file_path} not found"
            log_error(log_dir, 'file_not_found', error_message, process_id)
            continue
        with FileLock(cur_file_path + ".lock"):
            target_graph = ConjunctiveGraph()
            if zipped_output:
                with zipfile.ZipFile(cur_file_path, 'r') as zfile:
                    json_file_name = zfile.namelist()[0]
                    with zfile.open(json_file_name) as file:
                        file_content = file.read().decode('utf-8')
            else:
                with open(cur_file_path, 'r', encoding='utf-8') as file:
                    file_content = file.read()
            target_graph.parse(data=file_content, format='json-ld')
            update = False
            for subject in subjects:
                original_predicates = set(graph.predicates(subject, unique=True))
                target_predicates = set(target_graph.predicates(subject, unique=True))
                for predicate in original_predicates:
                    original_values = set(graph.objects(subject, predicate, unique=True))
                    target_values = set(target_graph.objects(subject, predicate, unique=True))
                    if original_values != target_values:
                        if predicate in target_predicates:
                            log_error(log_dir, 'different_predicate_value', f"Different values for predicate {predicate} for subject {subject} in {cur_file_path}", process_id)
                            context = target_graph.get_context(subject)
                            context.remove((subject, predicate, None))
                            for value in original_values:
                                context.add((subject, predicate, value))
                            update = True
                        else:
                            log_error(log_dir, 'missing_predicate', f"Predicate {predicate} missing for subject {subject} in {cur_file_path}", process_id)
            if update:
                if zipped_output:
                    with zipfile.ZipFile(cur_file_path, 'w') as zfile:
                        zfile.writestr(json_file_name, target_graph.serialize(format='json-ld', indent=None, ensure_ascii=False).encode('utf-8'))
                else:
                    with open(cur_file_path, 'w', encoding='utf-8') as file:
                        file.write(target_graph.serialize(format='json-ld', indent=None, ensure_ascii=False))
    if use_multiprocessing:
        queue.put(1)

def init_queue(q):
    global queue
    queue = q

def find_json_files(input_folder, exclude_dir='prov', zipped_input=None):
    """
    Recursively finds all files in the input folder, excluding those within directories named `exclude_dir`.
    """
    file_paths = []
    for root, dirs, files in os.walk(input_folder):
        if exclude_dir in root.split(os.sep):
            continue  # Skip files within directories named `exclude_dir`
        for file in files:
            if zipped_input and file.endswith('.zip'):
                file_paths.append(os.path.join(root, file))
            elif not zipped_input and file.endswith('.json'):
                file_paths.append(os.path.join(root, file))
    return file_paths

def main(input_folder, output_folder, use_multiprocessing, log_file_path, zipped_input, zipped_output):
    file_paths = find_json_files(input_folder, zipped_input=zipped_input)
    log_dir = log_file_path.rsplit('.', 1)[0] + "_logs"
    os.makedirs(log_dir, exist_ok=True)
    if use_multiprocessing:
        with Manager() as manager:
            queue = manager.Queue()
            with Pool(initializer=init_queue, initargs=(queue,)) as pool:
                for idx, filename in enumerate(file_paths):
                    file_path = os.path.join(input_folder, filename)
                    pool.apply_async(process_file, args=(file_path, output_folder, log_dir, zipped_input, zipped_output, idx, use_multiprocessing))
                pool.close()

                pbar = tqdm(total=len(file_paths))
                completed_tasks = 0
                while completed_tasks < len(file_paths):
                    queue.get()
                    completed_tasks += 1
                    pbar.update(1)
                pbar.close()

                pool.join()
    else:
        pbar = tqdm(file_paths, desc="Processing files", total=len(file_paths))
        for idx, filename in enumerate(file_paths):
            file_path = os.path.join(input_folder, filename)
            process_file(file_path, output_folder, log_dir, zipped_input, zipped_output, idx, use_multiprocessing)
            pbar.update()
        pbar.close()

    merge_logs(log_dir, log_file_path)
    shutil.rmtree(log_dir, ignore_errors=True)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process RDF data from JSONLD.GZ files.")
    parser.add_argument("input_folder", type=str, help="The folder containing .jsonld.gz files")
    parser.add_argument("output_folder", type=str, help="The folder to find matching files based on entities")
    parser.add_argument("--multiprocessing", action="store_true", help="Enable multiprocessing to process files")
    parser.add_argument("--log_file", type=str, required=True, help="The JSON file to log errors")
    parser.add_argument("--zipped_output", action="store_true", help="Indicate if the output files are zipped")
    parser.add_argument("--zipped_input", action="store_true", help="Indicate if the input files are zipped")

    args = parser.parse_args()

    main(args.input_folder, args.output_folder, args.multiprocessing, args.log_file, args.zipped_input, args.zipped_output)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from __future__ import annotations

import os
import re
from time import sleep, time
from typing import Set

import validators
import yaml
from oc_ocdm import Storer
from oc_ocdm.counter_handler.redis_counter_handler import RedisCounterHandler
from oc_ocdm.graph import GraphSet
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.prov import ProvSet
from oc_ocdm.reader import Reader
from oc_ocdm.support.support import build_graph_from_results
from rdflib import RDF, ConjunctiveGraph, URIRef
from requests.exceptions import RequestException
from SPARQLWrapper import JSON, SPARQLWrapper


class EntityCache:
    def __init__(self):
        self.cache: Set[URIRef] = set()
    
    def add(self, entity: URIRef) -> None:
        """Add an entity to the cache"""
        self.cache.add(entity)
    
    def is_cached(self, entity: URIRef) -> bool:
        """Check if an entity is in the cache"""
        return entity in self.cache
    
    def clear(self) -> None:
        """Clear all cached entities"""
        self.cache.clear()


class MetaEditor:
    property_to_remove_method = {
        'http://purl.org/spar/datacite/hasIdentifier': 'remove_identifier',
        'http://purl.org/spar/pro/isHeldBy': 'remove_is_held_by',
        'http://purl.org/vocab/frbr/core#embodiment': 'remove_format',
        'http://purl.org/spar/pro/isDocumentContextFor': 'remove_is_held_by',
        'https://w3id.org/oc/ontology/hasNext': 'remove_next'
    }

    def __init__(self, meta_config: str, resp_agent: str, save_queries: bool = False):
        with open(meta_config, encoding='utf-8') as file:
            settings = yaml.full_load(file)
        self.endpoint = settings['triplestore_url']
        self.base_dir = os.path.join(settings['output_rdf_dir'], 'rdf') + os.sep
        self.base_iri = settings['base_iri']
        self.resp_agent = resp_agent
        self.dir_split = settings['dir_split_number']
        self.n_file_item = settings['items_per_file']
        self.zip_output_rdf = settings['zip_output_rdf']
        self.reader = Reader()
        self.save_queries = save_queries
        self.update_queries = []
        
        # Redis configuration
        self.redis_host = settings.get('redis_host', 'localhost')
        self.redis_port = settings.get('redis_port', 6379)
        self.redis_db = settings.get('redis_db', 5)
        self.counter_handler = RedisCounterHandler(host=self.redis_host, port=self.redis_port, db=self.redis_db)

        self.entity_cache = EntityCache()
        self.relationship_cache = {}

    def make_sparql_query_with_retry(self, sparql: SPARQLWrapper, query, max_retries=5, backoff_factor=0.3):
        sparql.setQuery(query)
        sparql.setReturnFormat(JSON)

        start_time = time()
        for attempt in range(max_retries):
            try:
                return sparql.queryAndConvert()
            except Exception as e:
                duration = time() - start_time
                if attempt < max_retries - 1:
                    sleep_time = backoff_factor * (2 ** attempt)
                    print(
                        query, 
                        duration, 
                        f'retry_{attempt + 1}'
                    )
                    sleep(sleep_time)
                else:
                    print(f"SPARQL query failed after {max_retries} attempts: {e}")
                    raise

    def update_property(self, res: URIRef, property: str, new_value: str|URIRef) -> None:
        supplier_prefix = self.__get_supplier_prefix(res)
        g_set = GraphSet(self.base_iri, supplier_prefix=supplier_prefix, custom_counter_handler=self.counter_handler)
        self.reader.import_entity_from_triplestore(g_set, self.endpoint, res, self.resp_agent, enable_validation=False)
        if validators.url(new_value):
            new_value = URIRef(new_value)
            self.reader.import_entity_from_triplestore(g_set, self.endpoint, new_value, self.resp_agent, enable_validation=False)
            getattr(g_set.get_entity(res), property)(g_set.get_entity(new_value))
        else:
            getattr(g_set.get_entity(res), property)(new_value)
        self.save(g_set, supplier_prefix)
    
    def delete(self, res: str, property: str = None, object: str = None) -> None:
        supplier_prefix = self.__get_supplier_prefix(res)
        g_set = GraphSet(self.base_iri, supplier_prefix=supplier_prefix, custom_counter_handler=self.counter_handler)
        try:
            self.reader.import_entity_from_triplestore(g_set, self.endpoint, res, self.resp_agent, enable_validation=False)
        except ValueError as e:
            print(f"ValueError for entity {res}: {e}")
            inferred_type = self.infer_type_from_uri(res)
            if inferred_type:
                print(f"Inferred type {inferred_type} for entity {res}")
                sparql: SPARQLWrapper = SPARQLWrapper(self.endpoint)
                query: str = f"SELECT ?s ?p ?o WHERE {{BIND (<{res}> AS ?s). ?s ?p ?o.}}"
                sparql.setQuery(query)
                sparql.setMethod('GET')
                sparql.setReturnFormat(JSON)
                result = sparql.queryAndConvert()['results']['bindings']
                preexisting_graph: ConjunctiveGraph = build_graph_from_results(result)
                self.add_entity_with_type(g_set, res, inferred_type, preexisting_graph)
            else:
                return
        if not g_set.get_entity(URIRef(res)):
            return
        if property:
            remove_method = self.property_to_remove_method[property] if property in self.property_to_remove_method else property.replace('has', 'remove') if property.startswith('has') else f'remove_{property}'
            if object:
                if validators.url(object):
                    self.reader.import_entity_from_triplestore(g_set, self.endpoint, object, self.resp_agent, enable_validation=False)
                    # try:
                    getattr(g_set.get_entity(URIRef(res)), remove_method)(g_set.get_entity(URIRef(object)))
                    # TypeError: AgentRole.remove_is_held_by() takes 1 positional argument but 2 were given
                    # except TypeError:
                    #     getattr(g_set.get_entity(URIRef(res)), remove_method)()
                else:
                    getattr(g_set.get_entity(URIRef(res)), remove_method)(object)
            else:
                getattr(g_set.get_entity(URIRef(res)), remove_method)()
        else:
            sparql = SPARQLWrapper(endpoint=self.endpoint)
            query = f'SELECT ?s WHERE {{?s ?p <{res}>.}}'       
            sparql.setQuery(query)
            sparql.setReturnFormat(JSON)
            result = sparql.queryAndConvert()
            for entity in result['results']['bindings']:
                self.reader.import_entity_from_triplestore(g_set, self.endpoint, URIRef(entity['s']['value']), self.resp_agent, enable_validation=False)
            entity_to_purge = g_set.get_entity(URIRef(res))
            entity_to_purge.mark_as_to_be_deleted()
        self.save(g_set, supplier_prefix)

    def merge(self, g_set: GraphSet, res: URIRef, other: URIRef) -> None:
        """
        Merge two entities and their related entities using batch import with caching.
        
        Args:
            g_set: The GraphSet containing the entities
            res: The main entity that will absorb the other
            other: The entity to be merged into the main one
        """
        # First get all related entities with a single SPARQL query
        related_entities = set()
        if other in self.relationship_cache:
            related_entities.update(self.relationship_cache[other])
        else:
            sparql = SPARQLWrapper(endpoint=self.endpoint)
            query = f'''
                PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                PREFIX datacite: <http://purl.org/spar/datacite/>
                PREFIX pro: <http://purl.org/spar/pro/>
                SELECT DISTINCT ?entity WHERE {{
                    {{?entity ?p <{other}>}} UNION 
                    {{<{other}> ?p ?entity}}
                    FILTER (?p != rdf:type) 
                    FILTER (?p != datacite:usesIdentifierScheme) 
                    FILTER (?p != pro:withRole)
                }}'''
            
            data = self.make_sparql_query_with_retry(sparql, query)
            other_related = {URIRef(result["entity"]["value"]) 
                            for result in data["results"]["bindings"] 
                            if result['entity']['type'] == 'uri'}
            
            self.relationship_cache[other] = other_related
            related_entities.update(other_related)
        if res in self.relationship_cache:
            related_entities.update(self.relationship_cache[res])
        else:
            # Query only for objects of the surviving entity if not in cache
            sparql = SPARQLWrapper(endpoint=self.endpoint)
            query = f'''
                PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                PREFIX datacite: <http://purl.org/spar/datacite/>
                PREFIX pro: <http://purl.org/spar/pro/>
                SELECT DISTINCT ?entity WHERE {{
                    <{res}> ?p ?entity
                    FILTER (?p != rdf:type) 
                    FILTER (?p != datacite:usesIdentifierScheme) 
                    FILTER (?p != pro:withRole)
                }}'''
            
            data = self.make_sparql_query_with_retry(sparql, query)
            res_related = {URIRef(result["entity"]["value"]) 
                        for result in data["results"]["bindings"] 
                        if result['entity']['type'] == 'uri'}
            
            self.relationship_cache[res] = res_related
            related_entities.update(res_related)
                
        entities_to_import = set([res, other])
        entities_to_import.update(related_entities)
        entities_to_import = {e for e in entities_to_import 
                            if not self.entity_cache.is_cached(e)}
        # Import only non-cached entities if there are any
        if entities_to_import:
            try:
                self.reader.import_entities_from_triplestore(
                    g_set=g_set,
                    ts_url=self.endpoint,
                    entities=list(entities_to_import),
                    resp_agent=self.resp_agent,
                    enable_validation=False,
                    batch_size=10
                )
                
                # Add newly imported entities to cache
                for entity in entities_to_import:
                    self.entity_cache.add(entity)
                    
            except ValueError as e:
                print(f"Error importing entities: {e}")
                return
            
        # Perform the merge
        res_as_entity = g_set.get_entity(res)
        other_as_entity = g_set.get_entity(other)
        
        is_both_expression = all(
            GraphEntity.iri_expression in entity.g.objects(entity.res, RDF.type)
            for entity in [res_as_entity, other_as_entity]
        )
        
        if is_both_expression:
            res_as_entity.merge(other_as_entity, prefer_self=True)
        else:
            res_as_entity.merge(other_as_entity)

    def sync_rdf_with_triplestore(self, res: str, source_uri: str = None) -> bool:
        supplier_prefix = self.__get_supplier_prefix(res)
        g_set = GraphSet(self.base_iri, supplier_prefix=supplier_prefix, custom_counter_handler=self.counter_handler)
        try:
            self.reader.import_entity_from_triplestore(g_set, self.endpoint, res, self.resp_agent, enable_validation=False)
            self.save(g_set, supplier_prefix)
            return True
        except ValueError:
            try:
                self.reader.import_entity_from_triplestore(g_set, self.endpoint, source_uri, self.resp_agent, enable_validation=False)
            except ValueError:
                res_filepath = self.find_file(self.base_dir, self.dir_split, self.n_file_item, source_uri, self.zip_output_rdf)
                if not res_filepath:
                    return False
                imported_graph = self.reader.load(res_filepath)
                self.reader.import_entities_from_graph(g_set, imported_graph, self.resp_agent)
                res_entity = g_set.get_entity(URIRef(source_uri))
                if res_entity:
                    for res, entity in g_set.res_to_entity.items():
                        triples_list = list(entity.g.triples((URIRef(source_uri), None, None)))
                        for triple in triples_list:
                            entity.g.remove(triple)
                    self.save(g_set, supplier_prefix)
                return False
            
    def save(self, g_set: GraphSet, supplier_prefix: str = "") -> None:
        provset = ProvSet(g_set, self.base_iri, wanted_label=False, 
                         supplier_prefix=supplier_prefix, 
                         custom_counter_handler=self.counter_handler)
        provset.generate_provenance()
        graph_storer = Storer(g_set, dir_split=self.dir_split, 
                             n_file_item=self.n_file_item, 
                             zip_output=self.zip_output_rdf)
        prov_storer = Storer(provset, dir_split=self.dir_split, 
                            n_file_item=self.n_file_item, 
                            zip_output=self.zip_output_rdf)
        graph_storer.store_all(self.base_dir, self.base_iri)
        prov_storer.store_all(self.base_dir, self.base_iri)
        graph_storer.upload_all(self.endpoint, base_dir=self.base_dir, 
                              save_queries=self.save_queries)
        g_set.commit_changes()
        
    def __get_supplier_prefix(self, uri: str) -> str:
        entity_regex: str = r'^(.+)/([a-z][a-z])/(0[1-9]+0)?([1-9][0-9]*)$'
        entity_match = re.match(entity_regex, uri)
        return entity_match.group(3)

    def find_file(self, rdf_dir: str, dir_split_number: str, items_per_file: str, uri: str, zip_output_rdf: bool) -> str|None:
        entity_regex: str = r'^(https:\/\/w3id\.org\/oc\/meta)\/([a-z][a-z])\/(0[1-9]+0)?([1-9][0-9]*)$'
        entity_match = re.match(entity_regex, uri)
        if entity_match:
            cur_number = int(entity_match.group(4))
            cur_file_split: int = 0
            while True:
                if cur_number > cur_file_split:
                    cur_file_split += items_per_file
                else:
                    break
            cur_split: int = 0
            while True:
                if cur_number > cur_split:
                    cur_split += dir_split_number
                else:
                    break
            short_name = entity_match.group(2)
            sub_folder = entity_match.group(3)
            cur_dir_path = os.path.join(rdf_dir, short_name, sub_folder, str(cur_split))
            extension = '.zip' if zip_output_rdf else '.json'
            cur_file_path = os.path.join(cur_dir_path, str(cur_file_split)) + extension
            return cur_file_path

    def infer_type_from_uri(self, uri: str) -> str:
        if os.path.join(self.base_iri, 'br') in uri:
            return GraphEntity.iri_expression
        elif os.path.join(self.base_iri, 'ar') in uri:
            return GraphEntity.iri_role_in_time
        elif os.path.join(self.base_iri, 'ra') in uri:
            return GraphEntity.iri_agent
        elif os.path.join(self.base_iri, 're') in uri:
            return GraphEntity.iri_manifestation
        elif os.path.join(self.base_iri, 'id') in uri:
            return GraphEntity.iri_identifier
        return None

    def add_entity_with_type(self, g_set: GraphSet, res: str, entity_type: str, preexisting_graph: ConjunctiveGraph):
        subject = URIRef(res)
        if entity_type == GraphEntity.iri_expression:
            g_set.add_br(resp_agent=self.resp_agent, res=subject, preexisting_graph=preexisting_graph)
        elif entity_type == GraphEntity.iri_role_in_time:
            g_set.add_ar(resp_agent=self.resp_agent, res=subject, preexisting_graph=preexisting_graph)
        elif entity_type == GraphEntity.iri_agent:
            g_set.add_ra(resp_agent=self.resp_agent, res=subject, preexisting_graph=preexisting_graph)
        elif entity_type == GraphEntity.iri_manifestation:
            g_set.add_re(resp_agent=self.resp_agent, res=subject, preexisting_graph=preexisting_graph)
        elif entity_type == GraphEntity.iri_identifier:
            g_set.add_id(resp_agent=self.resp_agent, res=subject, preexisting_graph=preexisting_graph)


import os
import json
import redis
from tqdm import tqdm
from multiprocessing import Pool, cpu_count


def load_json_to_redis(file_path, redis_host='localhost', redis_port=6379, db_number=14, omid_db_number=10):
    r = redis.Redis(host=redis_host, port=redis_port, db=db_number)
    r_omid = redis.Redis(host=redis_host, port=redis_port, db=omid_db_number)

    with open(file_path, 'r', encoding='utf-8') as file:
        doi_data = {}
        for line in file:
            item = json.loads(line)
            doi = item.get('id', '')
            omid_key = f'doi:{doi.lower()}'
            omid = r_omid.get(omid_key)
            if omid:
                omid_decoded = omid.decode()
                omid_record_key = f'omid:{omid_decoded}'
                if not r.exists(omid_record_key):
                    doi_data[omid_record_key] = 'https://archive.org/details/datacite_dump_20211022'
                    if len(doi_data) >= 1000:
                        r.mset(doi_data)
                        doi_data = {}
        if doi_data:
            r.mset(doi_data)
    print("Caricamento completato.")

folder_path = '/srv/data/arcangelo/datacite_dump_20211022/datacite_dump_20211022.json'
load_json_to_redis(folder_path)


import argparse
import csv
import os
import re

import psycopg2


def get_csv_data(filepath:str):
    if not os.path.splitext(filepath)[1].endswith('.csv'):
        return list()
    field_size_changed = False
    cur_field_size = 128
    data = list()
    while not data:
        try:
            with open(filepath, 'r', encoding='utf8') as f:
                data = list(csv.DictReader(f, delimiter=','))
        except csv.Error:
            cur_field_size *= 2
            csv.field_size_limit(cur_field_size)
            field_size_changed = True
    if field_size_changed:
        csv.field_size_limit(128)
    return data

def extract_omids(text, field, regex_pattern):
    identifiers = []
    if field == 'id':
        identifiers.extend([identifier for identifier in text.split() if not identifier.startswith('omid:')])
    elif field in {'author', 'editor', 'publisher'}:
        agents = text.split('; ')
        for agent in agents:
            if re.match(regex_pattern, agent):
                identifiers.extend([identifier for identifier in re.match(regex_pattern, agent).group(2).split() if not identifier.startswith('omid:')])
    elif field in 'venue':
        if re.match(regex_pattern, text):
            identifiers.extend([identifier for identifier in re.match(regex_pattern, text).group(2).split() if not identifier.startswith('omid:')])
    return identifiers

def save_to_database(omids, source, connection):
    with connection.cursor() as cursor:
        for omid in omids:
            cursor.execute("INSERT INTO omid_table (omid, source) VALUES (%s, %s)", (omid, source))
    connection.commit()

def main():
    parser = argparse.ArgumentParser(description='Process CSV files in a directory.')
    parser.add_argument('directory', type=str, help='Directory containing CSV files')
    args = parser.parse_args()

    regex_pattern = r'\s*(.*?)\s*\[\s*((?:[^\s]+:[^\s]+)?(?:\s+[^\s]+:[^\s]+)*)\s*\]'
    connection = psycopg2.connect("dbname='oc_meta' user='postgres' password='Permesiva1!'")

    for filename in os.listdir(args.directory):
        if filename.endswith('.csv'):
            reader = get_csv_data(os.path.join(args.directory, filename))
            for row in reader:
                for field in ['id', 'author', 'editor', 'publisher', 'venue']:
                    omids = extract_omids(row[field], field, regex_pattern)
                    print(omids)
                    # save_to_database(omids, "Crossref", connection)

    connection.close()

if __name__ == "__main__":
    main()


import argparse
import json
from oc_meta.plugins.editor import MetaEditor
from tqdm import tqdm
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from SPARQLWrapper import SPARQLWrapper, JSON
from yaml import safe_load

def query_and_delete_triples(uri, meta_config, resp_agent, stop_file, endpoint):
    if os.path.exists(stop_file):
        return "stopped"

    meta_editor = MetaEditor(meta_config=meta_config, resp_agent=resp_agent)
    
    # Query to find triples where the entity is an object
    query = f"""
    SELECT ?s ?p
    WHERE {{
        ?s ?p <{uri}> .
    }}
    """
    ts = SPARQLWrapper(endpoint)
    ts.setQuery(query)
    ts.setReturnFormat(JSON)
    results = ts.queryAndConvert()
    # Delete each triple
    for result in results['results']['bindings']:
        s = result['s']['value']
        p = result['p']['value']
        if p not in {'http://www.w3.org/1999/02/22-rdf-syntax-ns#type', 'http://purl.org/spar/datacite/usesIdentifierScheme', 'http://purl.org/spar/pro/withRole'}:
            meta_editor.delete(res=s, property=p, object=uri)
    
    return "deleted"

def single_process_deletion(uris, meta_config, resp_agent, stop_file, endpoint):    
    with tqdm(total=len(uris), desc="Deleting triples for URIs") as pbar:
        for uri in uris:
            if os.path.exists(stop_file):
                print("Stop file detected. Halting further deletions.")
                break
            query_and_delete_triples(uri, meta_config, resp_agent, stop_file, endpoint)
            pbar.update(1)

def multi_process_deletion(uris, meta_config, resp_agent, stop_file, endpoint):
    with ProcessPoolExecutor() as executor:
        futures = {executor.submit(query_and_delete_triples, uri, meta_config, resp_agent, stop_file, endpoint): uri for uri in uris}
        
        with tqdm(total=len(uris), desc="Deleting triples for URIs") as pbar:
            for future in as_completed(futures):
                result = future.result()
                if result == "stopped":
                    print("Stop file detected. Halting further deletions.")
                    break
                pbar.update(1)

def main():
    parser = argparse.ArgumentParser(description="Delete triples for URIs using MetaEditor.")
    parser.add_argument("json_file", type=str, help="Path to the JSON file containing URIs")
    parser.add_argument("meta_config", type=str, help="Path to the MetaEditor configuration file")
    parser.add_argument("resp_agent", type=str, help="Responsible agent URI")
    parser.add_argument("--stop_file", type=str, default=".stop_deletions", help="Path to the stop file for graceful termination")
    parser.add_argument("--multiprocessing", action="store_true", help="Use multiprocessing for deletions")

    args = parser.parse_args()

    if os.path.exists(args.stop_file):
        os.remove(args.stop_file)

    with open(args.json_file, 'r') as file:
        uris = json.load(file)

    with open(args.meta_config, 'r', encoding='utf8') as f:
        meta_config = safe_load(f)
    endpoint = meta_config['triplestore_url']

    if args.multiprocessing:
        multi_process_deletion(uris, args.meta_config, args.resp_agent, args.stop_file, endpoint)
    else:
        single_process_deletion(uris, args.meta_config, args.resp_agent, args.stop_file, endpoint)

if __name__ == "__main__":
    main()


import argparse
import gzip
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from tqdm import tqdm

def search_in_file(file_path, search_string):
    try:
        with gzip.open(file_path, 'rt', encoding='utf-8') as f:
            for line_number, line in enumerate(f, 1):
                if search_string in line:
                    return file_path, line_number, line.strip()
    except Exception as e:
        print(f"Errore durante la lettura del file {file_path}: {e}")
    return None

def main():
    parser = argparse.ArgumentParser(description="Cerca una stringa in file .nq.gz")
    parser.add_argument("folder", help="Percorso della cartella contenente i file .nq.gz")
    parser.add_argument("search_string", help="Stringa da cercare nei file")
    args = parser.parse_args()

    folder_path = Path(args.folder)
    search_string = args.search_string

    if not folder_path.is_dir():
        print(f"Il percorso specificato non è una cartella valida: {folder_path}")
        return

    nq_files = list(folder_path.glob("*.nq.gz"))

    if not nq_files:
        print(f"Nessun file .nq.gz trovato nella cartella: {folder_path}")
        return

    print(f"Ricerca di '{search_string}' in {len(nq_files)} file...")

    results = []
    with ProcessPoolExecutor() as executor:
        futures = [executor.submit(search_in_file, file, search_string) for file in nq_files]
        for future in tqdm(as_completed(futures), total=len(nq_files), desc="Progresso", unit="file"):
            result = future.result()
            if result:
                results.append(result)

    if results:
        print(f"\nLa stringa '{search_string}' è stata trovata nei seguenti file:")
        for file_path, line_number, line in results:
            print(f"\nFile: {file_path}")
            print(f"Riga numero: {line_number}")
            print(f"Contenuto: {line}")
    else:
        print(f"\nLa stringa '{search_string}' non è stata trovata in nessun file.")

if __name__ == "__main__":
    main()


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE

import argparse
import json
import os
import zipfile
from datetime import datetime
from concurrent.futures import ProcessPoolExecutor, as_completed
from pebble import ProcessPool
import rdflib
from rdflib import Namespace, Literal, URIRef, ConjunctiveGraph
from rdflib.namespace import RDF, XSD
import tempfile

from tqdm import tqdm


DC = Namespace("http://purl.org/dc/terms/")
PROV = Namespace("http://www.w3.org/ns/prov#")
RESP_AGENT = URIRef("https://w3id.org/oc/meta/prov/pa/1")

def process_zip_file(zip_path):
    output_dir = os.path.join(os.path.dirname(zip_path), os.path.splitext(os.path.basename(zip_path))[0], 'prov')
    output_zip_path = os.path.join(output_dir, 'se.zip')

    if os.path.exists(output_zip_path):
        return

    g = ConjunctiveGraph()
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        for file_name in zip_ref.namelist():
            with zip_ref.open(file_name) as file:
                json_data = json.load(file)
                for graph in json_data:
                    for entity in graph["@graph"]:
                        entity_id = URIRef(entity["@id"])
                        graph_uri = URIRef(f'{entity_id}/prov/')
                        se_uri = URIRef(f"{entity_id}/prov/se/1")
                        g.add((se_uri, RDF.type, PROV.Entity, graph_uri))
                        g.add((se_uri, DC.description, Literal(f"The entity '{entity_id}' has been created."), graph_uri))
                        g.add((se_uri, PROV.generatedAtTime, Literal(datetime.utcnow(), datatype=XSD.dateTime), graph_uri))
                        g.add((se_uri, PROV.specializationOf, entity_id, graph_uri))
                        g.add((se_uri, PROV.wasAttributedTo, RESP_AGENT, graph_uri))
    graph_jsonld = g.serialize(format='json-ld', ensure_ascii=False, indent=None, encoding='utf8')
    
    # Create the directory if it doesn't exist
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    with zipfile.ZipFile(output_zip_path, 'w') as output_zip:
        output_zip.writestr('se.json', graph_jsonld)

def find_zip_files_in_subdir(subdir, process_immediately=False):
    if process_immediately:
        if os.path.exists(subdir):
            for root, dirs, files in os.walk(subdir):
                if 'prov' not in root.split(os.sep):
                    for file in files:
                        if file.endswith('.zip'):
                            process_zip_file(os.path.join(root, file))
    else:
        zip_files = []
        if os.path.exists(subdir):
            for root, dirs, files in os.walk(subdir):
                if 'prov' not in root.split(os.sep):
                    for file in files:
                        if file.endswith('.zip'):
                            zip_files.append(os.path.join(root, file))
        return zip_files

def find_subdirs(directory):
    subdirs = []
    for dir in os.listdir(directory):
        subdirs.append(os.path.join(directory, dir))
    return subdirs

def search_zip_files(directory, show_progress):
    main_subdirs = ['ar', 'br', 'ra', 're', 'id']
    subdirs = []

    for main_subdir in main_subdirs:
        main_subdir_path = os.path.join(directory, main_subdir)
        subdirs.extend(find_subdirs(main_subdir_path))

    if show_progress:
        zip_files = []
        with ProcessPoolExecutor() as executor:
            futures = [executor.submit(find_zip_files_in_subdir, subdir) for subdir in subdirs]
            for future in futures:
                zip_files.extend(future.result())
        with ProcessPoolExecutor() as executor:
            futures = {executor.submit(process_zip_file, file): file for file in zip_files}
            for future in tqdm(as_completed(futures), total=len(futures)):
                pass
    else:
        for subdir in subdirs:
            find_zip_files_in_subdir(subdir, process_immediately=True)

def main():
    parser = argparse.ArgumentParser(description="Process JSON-LD files within ZIP archives")
    parser.add_argument("directory", help="Directory to search for ZIP files")
    parser.add_argument("--progress", help="Show progress bar", action="store_true")
    args = parser.parse_args()
    search_zip_files(args.directory, args.progress)

if __name__ == "__main__":
    main()


import argparse
import os
from SPARQLWrapper import SPARQLWrapper, JSON
from oc_meta.plugins.editor import MetaEditor
from rdflib import URIRef
from tqdm import tqdm

def query_publishers(endpoint_url, output_file):
    if os.path.isfile(output_file):
        with open(output_file, 'r') as file:
            publishers = [line.strip() for line in file]
    else:
        sparql_query = """
        PREFIX foaf: <http://xmlns.com/foaf/0.1/>
        SELECT DISTINCT ?pub
        WHERE {
            ?pub a foaf:Agent;
                foaf:name ?name.
            FILTER CONTAINS(?name, ";").
        }
        """

        sparql = SPARQLWrapper(endpoint_url)
        sparql.setQuery(sparql_query)
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()

        publishers = [result["pub"]["value"] for result in results["results"]["bindings"]]
        with open(output_file, 'w') as file:
            for pub in publishers:
                print(pub)
                file.write(pub + '\n')
    return publishers

def update_publishers_names(endpoint_url, publishers, config_path, resp_agent):
    meta_editor = MetaEditor(meta_config=config_path, resp_agent=URIRef(resp_agent))
    sparql = SPARQLWrapper(endpoint_url)
    sparql.setReturnFormat(JSON)

    for pub in tqdm(publishers, desc="Updating publisher names"):
        sparql.setQuery(f"""
            SELECT ?old_value
            WHERE {{
                <{pub}> foaf:name ?old_value.
            }}""")
        results = sparql.query().convert()
        old_value = results["results"]["bindings"][0]["old_value"]["value"]
        new_value = old_value.replace(";", "")
        meta_editor.update_property(res=URIRef(pub), property="has_name", new_value=new_value)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Query SPARQL endpoint for publishers whose name contains a semicolon, work with MetaEditor on the URIs.")
    parser.add_argument("endpoint", help="SPARQL endpoint URL")
    parser.add_argument("output_file", help="Path to the output file where URIs will be saved")
    parser.add_argument("meta_config_path", help="Path to the MetaEditor configuration file")
    parser.add_argument("resp_agent", help="Responsible agent for MetaEditor operations")
    args = parser.parse_args()

    publishers = query_publishers(args.endpoint, args.output_file)
    update_publishers_names(args.endpoint, publishers, args.meta_config_path, args.resp_agent)


import argparse
import zipfile
import os
import json
from rdflib import ConjunctiveGraph, URIRef
from tqdm import tqdm
from multiprocessing import Pool, cpu_count
from oc_meta.run.align_rdf_with_triplestore import find_paths
from functools import partial

BASE_IRI = "https://w3id.org/oc/meta/"
DIR_SPLIT = 10000
N_FILE_ITEM = 1000
IS_JSON = True

def extract_and_process_json(zip_path, source_folder):
    processed_zip_cache = {}
    changes_made = False
    with zipfile.ZipFile(zip_path, 'r') as z:
        for file_info in z.infolist():
            if file_info.filename.endswith('.json'):
                with z.open(file_info.filename) as file:
                    json_data = json.load(file)
                    graph = ConjunctiveGraph()
                    graph.parse(data=json.dumps(json_data), format='json-ld')
                    for subject in graph.subjects(unique=True):
                        current_context = URIRef(str(subject).split('/prov/')[0] + '/prov/')
                        if not (subject, URIRef("http://www.w3.org/ns/prov#hadPrimarySource"), None) in graph:
                            _, cur_file_path = find_paths(subject, source_folder, BASE_IRI, "_", DIR_SPLIT, N_FILE_ITEM, IS_JSON)
                            cur_file_path = cur_file_path.replace('.json', '.zip')
                            if cur_file_path not in processed_zip_cache:
                                if os.path.exists(cur_file_path):
                                    with zipfile.ZipFile(cur_file_path, 'r') as cur_zip:
                                        for cur_file_info in cur_zip.infolist():
                                            if cur_file_info.filename.endswith('.json'):
                                                with cur_zip.open(cur_file_info.filename) as cur_file:
                                                    cur_json_data = json.load(cur_file)
                                                    cur_graph = ConjunctiveGraph()
                                                    cur_graph.parse(data=json.dumps(cur_json_data), format='json-ld')
                                                    processed_zip_cache[cur_file_path] = cur_graph
                            if cur_file_path in processed_zip_cache:
                                cur_graph: ConjunctiveGraph = processed_zip_cache[cur_file_path]
                                # Check for the prov#hadPrimarySource predicate
                                for o in cur_graph.objects(subject, URIRef("http://www.w3.org/ns/prov#hadPrimarySource"), unique=True):
                                    if str(o) == 'https://api.crossref.org/':
                                        graph.add((subject, URIRef("http://www.w3.org/ns/prov#hadPrimarySource"), URIRef('https://api.crossref.org/snapshots/monthly/2022/12/all.json.tar.gz'), current_context))
                                        changes_made = True
                                    elif str(o) == 'https://api.datacite.org/':
                                        graph.add((subject, URIRef("http://www.w3.org/ns/prov#hadPrimarySource"), URIRef('https://archive.org/details/datacite_dump_20211022'), current_context))
                                        changes_made = True
                                    elif str(o) == 'https://nih.figshare.com/collections/iCite_Database_Snapshots_NIH_Open_Citation_Collection_/4586573/42':
                                        graph.add((subject, URIRef("http://www.w3.org/ns/prov#hadPrimarySource"), URIRef('https://nih.figshare.com/collections/iCite_Database_Snapshots_NIH_Open_Citation_Collection_/4586573/36'), current_context))
                                        changes_made = True
                                    elif str(o) == 'https://api.crossref.org/snapshots/monthly/2023/09/all.json.tar.gz':
                                        graph.add((subject, URIRef("http://www.w3.org/ns/prov#hadPrimarySource"), URIRef('https://api.crossref.org/snapshots/monthly/2023/09/all.json.tar.gz'), current_context))
                                        changes_made = True
                                    elif str(o) == 'https://doi.org/10.5281/zenodo.7845968':
                                        graph.add((subject, URIRef("http://www.w3.org/ns/prov#hadPrimarySource"), URIRef('https://doi.org/10.5281/zenodo.7845968'), current_context))
                                        changes_made = True
                                    else:
                                        print(o)
    if changes_made:
        updated_json = graph.serialize(format='json-ld', indent=None, encoding='utf-8', ensure_ascii=False)
        with zipfile.ZipFile(zip_path, 'w', compression=zipfile.ZIP_DEFLATED, allowZip64=True) as z:
            z.writestr('se.json', updated_json)

def process_zip_files(zip_files, source_folder):
    func = partial(extract_and_process_json, source_folder=source_folder)
    with Pool(processes=cpu_count()) as pool:
        list(tqdm(pool.imap(func, zip_files), total=len(zip_files), desc="Processing se.zip files"))

def find_zip_files(destination_folder, source_folder):
    count_file = 'se_zip_file_paths.json'
    se_zip_files = []

    if os.path.exists(count_file):
        with open(count_file, 'r') as f:
            se_zip_files = json.load(f)
            if not se_zip_files:
                print("No se.zip files to process.")
                return
    else:
        for root, dirs, files in os.walk(destination_folder):
            for file in files:
                if file == 'se.zip':
                    se_zip_files.append(os.path.join(root, file))
        
        with open(count_file, 'w') as f:
            json.dump(se_zip_files, f)

    if se_zip_files:
        process_zip_files(se_zip_files, source_folder)
    
    if os.path.exists(count_file):
        os.remove(count_file)

def main():
    parser = argparse.ArgumentParser(description="Process se.zip files containing JSON-LD with RDF data.")
    parser.add_argument("destination_folder", type=str, help="Folder where to find se.zip files.")
    parser.add_argument("source_folder", type=str, help="Folder to store outputs (not used in this script).")
    args = parser.parse_args()

    find_zip_files(args.destination_folder, args.source_folder)

if __name__ == "__main__":
    main()


import argparse
import os
import zipfile
import json
import csv
from tqdm import tqdm
import logging

logging.basicConfig(filename='error_log_find_duplicated_resources.txt', level=logging.ERROR, 
                    format='%(asctime)s - %(levelname)s - %(message)s')

class UnionFind:
    def __init__(self):
        self.parent = {}
        self.rank = {}

    def find(self, item):
        if item not in self.parent:
            self.parent[item] = item
            self.rank[item] = 0
            return item
        
        if self.parent[item] != item:
            self.parent[item] = self.find(self.parent[item])
        return self.parent[item]

    def union(self, x, y):
        xroot = self.find(x)
        yroot = self.find(y)

        if xroot == yroot:
            return

        if self.rank[xroot] < self.rank[yroot]:
            self.parent[xroot] = yroot
        elif self.rank[xroot] > self.rank[yroot]:
            self.parent[yroot] = xroot
        else:
            self.parent[yroot] = xroot
            self.rank[xroot] += 1

def read_and_analyze_zip_files(folder_path, csv_path, resource_type):
    resources = {}
    
    if resource_type in ['br', 'both']:
        br_folder_path = os.path.join(folder_path, 'br')
        process_folder(br_folder_path, resources, 'br')
    
    if resource_type in ['ra', 'both']:
        ra_folder_path = os.path.join(folder_path, 'ra')
        process_folder(ra_folder_path, resources, 'ra')

    save_duplicates_to_csv(resources, csv_path)

def process_folder(folder_path, resources, expected_type):
    if not os.path.exists(folder_path):
        logging.error(f"La sottocartella '{expected_type}' non esiste nel percorso: {folder_path}")
        return

    zip_files = get_zip_files(folder_path)

    for zip_path in tqdm(zip_files, desc=f"Analizzando i file ZIP in {expected_type}"):
        try:
            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                for zip_file in zip_ref.namelist():
                    try:
                        with zip_ref.open(zip_file) as json_file:
                            data = json.load(json_file)
                            analyze_json(data, resources, zip_path, zip_file, expected_type)
                    except json.JSONDecodeError:
                        logging.error(f"Errore nel parsing JSON del file {zip_file} in {zip_path}")
                    except Exception as e:
                        logging.error(f"Errore nell'elaborazione del file {zip_file} in {zip_path}: {str(e)}")
        except zipfile.BadZipFile:
            logging.error(f"File ZIP corrotto o non valido: {zip_path}")
        except Exception as e:
            logging.error(f"Errore nell'apertura del file ZIP {zip_path}: {str(e)}")

def get_zip_files(folder_path):
    zip_files = []
    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if file.endswith('.zip') and not file == 'se.zip':
                zip_files.append(os.path.join(root, file))
    return zip_files

def analyze_json(data, resources, zip_path, zip_file, expected_type):
    for graph in data:
        for entity in graph.get("@graph", []):
            try:
                entity_id = entity["@id"]
                entity_type = get_entity_type(entity)
                
                if entity_type is None:
                    print(f"Tipo non specificato per l'entità {entity_id} nel file {zip_file} all'interno di {zip_path}. Assumendo tipo {expected_type}.")
                    entity_type = expected_type
                
                if entity_type == expected_type:
                    identifiers = get_identifiers(entity)
                    
                    if entity_id not in resources:
                        resources[entity_id] = set()
                    resources[entity_id].update(identifiers)
            except KeyError as e:
                logging.error(f"Chiave mancante nell'entità {entity.get('@id', 'ID sconosciuto')} "
                              f"nel file {zip_file} all'interno di {zip_path}: {str(e)}")
            except Exception as e:
                logging.error(f"Errore nell'analisi dell'entità {entity.get('@id', 'ID sconosciuto')} "
                              f"nel file {zip_file} all'interno di {zip_path}: {str(e)}")

def get_entity_type(entity):
    if "http://purl.org/spar/fabio/Expression" in entity.get("@type", []):
        return 'br'
    elif "http://xmlns.com/foaf/0.1/Agent" in entity.get("@type", []):
        return 'ra'
    return None

def get_identifiers(entity):
    identifiers = []
    for identifier in entity.get("http://purl.org/spar/datacite/hasIdentifier", []):
        if isinstance(identifier, dict) and "@id" in identifier:
            identifiers.append(identifier["@id"])
    return identifiers

def save_duplicates_to_csv(resources, csv_path):
    try:
        with open(csv_path, mode='w', newline='', encoding='utf-8') as csv_file:
            csv_writer = csv.writer(csv_file)
            csv_writer.writerow(['surviving_entity', 'merged_entities'])

            duplicates = find_duplicates(resources)
            for group in duplicates:
                surviving_entity = group[0]
                merged_entities = '; '.join(group[1:])
                csv_writer.writerow([surviving_entity, merged_entities])
    except Exception as e:
        logging.error(f"Errore nel salvataggio del file CSV {csv_path}: {str(e)}")

def find_duplicates(resources):
    uf = UnionFind()
    
    # First, create sets of identifiers for each entity
    for entity, identifiers in resources.items():
        for identifier in identifiers:
            uf.union(entity, identifier)
    
    # Then, group entities by their representative
    groups = {}
    for entity in resources:
        rep = uf.find(entity)
        if rep not in groups:
            groups[rep] = []
        groups[rep].append(entity)
    
    # Filter out groups with only one entity
    return [sorted(group) for group in groups.values() if len(group) > 1]

def main():
    parser = argparse.ArgumentParser(description="Trova risorse duplicate in base ai loro ID.")
    parser.add_argument("folder_path", type=str, help="Percorso della cartella contenente le sottocartelle 'br' e 'ra'")
    parser.add_argument("csv_path", type=str, help="Percorso del file CSV per salvare i duplicati")
    parser.add_argument("resource_type", type=str, choices=['br', 'ra', 'both'], 
                        help="Tipo di risorsa da analizzare: 'br' per risorse bibliografiche, 'ra' per agenti responsabili, 'both' per entrambi")
    args = parser.parse_args()

    read_and_analyze_zip_files(args.folder_path, args.csv_path, args.resource_type)

if __name__ == "__main__":
    main()


import argparse
import csv
import os

from tqdm import tqdm


def get_entity_type(entity_url):
    parts = entity_url.split('/')
    if 'oc' in parts and 'meta' in parts:
        try:
            return parts[parts.index('meta') + 1]
        except IndexError:
            return None
    return None

def process_csv(csv_file, entities_types_total, entities_types_done):
    with open(csv_file, mode='r', encoding='utf-8') as csvfile:
        reader = csv.DictReader(csvfile)
        rows = list(reader)
        
        for row in rows:
            entities = [row['surviving_entity']] + row['merged_entities'].split('; ')
            for entity in entities:
                entity_type = get_entity_type(entity)
                if entity_type:
                    if entity_type not in entities_types_total:
                        entities_types_total[entity_type] = set()
                    entities_types_total[entity_type].add(entity)
                    
                    if 'Done' in row and row['Done'].strip().lower() == 'true':
                        if entity_type not in entities_types_done:
                            entities_types_done[entity_type] = set()
                        entities_types_done[entity_type].add(entity)

def main():
    parser = argparse.ArgumentParser(description="Conta le entità uniche per tipo, evitando duplicati tra file e cartella, con conteggio opzionale delle entità 'Done'.")
    parser.add_argument("folder_path", help="Percorso alla cartella contenente altri file CSV")
    args = parser.parse_args()

    entities_types_total = {}
    entities_types_done = {}

    if os.path.exists(args.folder_path):
        csv_files = [f for f in os.listdir(args.folder_path) if f.endswith('.csv')]
        for filename in tqdm(csv_files, desc="Processando i file", unit="file"):
            file_path = os.path.join(args.folder_path, filename)
            process_csv(file_path, entities_types_total, entities_types_done)

    print("\nRisultati:")
    for entity_type, entities in entities_types_total.items():
        total = len(entities)
        done = len(entities_types_done.get(entity_type, set()))
        percentage = (done / total * 100) if total > 0 else 0
        print(f"Tipo di entità: {entity_type}, Totale entità uniche: {total}, Con 'Done'=True: {done}, Percentuale completata: {percentage:.2f}%")

if __name__ == "__main__":
    main()


import argparse
import os
import hashlib
import json
import requests
import yaml
from requests.exceptions import HTTPError

# Endpoint base di Figshare
BASE_URL = 'https://api.figshare.com/v2/account/articles'
CHUNK_SIZE = 1048576

def get_file_check_data(file_name):
    with open(file_name, 'rb') as fin:
        md5 = hashlib.md5()
        size = 0
        data = fin.read(CHUNK_SIZE)  # circa 10MB
        while data:
            size += len(data)
            md5.update(data)
            data = fin.read(CHUNK_SIZE)
        return md5.hexdigest(), size

def issue_request(method, url, token, data=None, binary=False):
    headers = {'Authorization': 'token ' + token}
    if data is not None and not binary:
        data = json.dumps(data)
    response = requests.request(method, url, headers=headers, data=data)
    try:
        response.raise_for_status()
        try:
            data = json.loads(response.content)
        except ValueError:
            data = response.content
    except HTTPError as error:
        print(f'Caught an HTTPError: {str(error)}')
        print('Body:\n', response.text)
        raise
    return data

def upload_parts(file_info, file_path, token):
    url = file_info['upload_url']
    result = issue_request(method='GET', url=url, token=token)
    print('Uploading parts:')
    with open(file_path, 'rb') as fin:
        for part in result['parts']:
            upload_part(file_info, fin, part, token)

def upload_part(file_info, stream, part, token):
    udata = file_info.copy()
    udata.update(part)
    url = '{upload_url}/{partNo}'.format(**udata)
    stream.seek(part['startOffset'])
    data = stream.read(part['endOffset'] - part['startOffset'] + 1)
    issue_request(method='PUT', url=url, data=data, binary=True, token=token)
    print('  Uploaded part {partNo} from {startOffset} to {endOffset}'.format(**part))

def create_file(article_id, file_name, file_path, token):
    url = f"{BASE_URL}/{article_id}/files"
    headers = {'Authorization': f'token {token}'}
    md5, size = get_file_check_data(file_path)
    data = {'name': os.path.basename(file_name), 'md5': md5, 'size': size}
    post_response = requests.post(url, headers=headers, json=data)
    get_response = requests.get(post_response.json()['location'], headers=headers)
    return get_response.json()

def main(config_path):
    with open(config_path) as f:
        config = yaml.safe_load(f)

    token = config['TOKEN']
    article_id = config['ARTICLE_ID']
    files_to_upload = config['files_to_upload']

    for file_path in files_to_upload:
        file_name = os.path.basename(file_path)
        print(f"Creazione del file {file_name} in Figshare...")
        file_info = create_file(article_id, file_name, file_path, token)
        upload_parts(file_info, file_path, token)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Upload files to Figshare.")
    parser.add_argument('config', help="Path to the YAML configuration file.")
    args = parser.parse_args()
    main(args.config)


import os
import csv
import redis
from tqdm import tqdm

def process_file(file_path, r, r_omid):
    with open(file_path, 'r', encoding='utf-8') as file:
        reader = csv.DictReader(file)
        id_data = {}
        for row in reader:
            ids = row.get('id', '').split()
            for id in ids:
                omid = r_omid.get(id)
                if omid:
                    omid_decoded = omid.decode()
                    omid_record_key = f'omid:{omid_decoded}'
                    if not r.exists(omid_record_key):
                        id_data[omid_record_key] = 'https://api.japanlinkcenter.org/'
                        if len(id_data) >= 1000:
                            r.mset(id_data)
                            id_data = {}
        if id_data:
            r.mset(id_data)

def load_csv_to_redis(folder_path, redis_host='localhost', redis_port=6379, db_number=14, omid_db_number=10):
    r = redis.Redis(host=redis_host, port=redis_port, db=db_number)
    r_omid = redis.Redis(host=redis_host, port=redis_port, db=omid_db_number)
    file_paths = [os.path.join(folder_path, f) for f in os.listdir(folder_path) if f.endswith('.csv')]
    for file_path in file_paths:
        process_file(file_path, r, r_omid)
    print("Caricamento completato.")

folder_path = '/home/arcangelo/meta_input_joci'
load_csv_to_redis(folder_path)